 precision highp float;

 vec2 textureCoordinateFix; //used for fix error

 uniform vec2 posV[20];
 uniform float paramV[17];
 //common
 uniform float sin_t;
 uniform float cos_t;
 uniform vec3 angles; // angles.x: pitch, angles.y: yaw, angles.z: roll
 uniform vec2 size;
 varying vec2 textureCoordinate;

 uniform sampler2D inputImageTexture;

 const float kLipsEdgeWrinkleSizeToLipsSizeRatio = 0.23;
 const float kPi = 3.14159265358979;
 const float kLipsEdgeWrinklesRotationAngle = 0.025 * kPi;

 //common func
 float getEllipseMask(in vec2 sigmaScaling, in vec2 point,in vec2 center, in vec2 boundingSize, float a1, float a2) {
     // un-normalization
     //boundingSize = boundingSize * size;
     //point = point * size;
     //center = center * size;

     vec2 translate = (point - center);
     // rotate and scale
     vec2 v = vec2(dot(translate, vec2(cos_t, sin_t)), dot(translate, vec2(-sin_t, cos_t))) * sigmaScaling;
     float d2 = sqrt(dot(v / boundingSize, v / boundingSize));
     return 1.0 - smoothstep(a1, a2, d2);
 }

 float getEllipseMask2(in vec2 sigmaScaling, in vec2 point,in vec2 center, in vec2 boundingSize, float a1, float a2, float ang) {
     // un-normalization
     //boundingSize = boundingSize * size;
     //point = point * size;
     //center = center * size;

     float cos_t = cos(ang);
     float sin_t = sin(ang);
     vec2 translate = (point - center);
     // rotate and scale
     vec2 v = vec2(dot(translate, vec2(cos_t, sin_t)), dot(translate, vec2(-sin_t, cos_t))) * sigmaScaling;

     float d2 = sqrt(dot(v / boundingSize, v / boundingSize));

     return 1.0 - smoothstep(a1, a2, d2);
 }

 vec2 rotate2DPoint(vec2 point, float angle) {
     return vec2(cos(angle) * point.x - sin(angle) * point.y, sin(angle) * point.x + cos(angle) * point.y);
 }

 //for eye part
 float eye_getLeftEyeFineMask(vec2 sigmaScaling, float a1, float a2) {
     return getEllipseMask(sigmaScaling, textureCoordinateFix, posV[1], posV[0], a1, a2);
 }
 float eye_getRightEyeFineMask(vec2 sigmaScaling, float a1, float a2) {
     return getEllipseMask(sigmaScaling, textureCoordinateFix, posV[3], posV[2], a1, a2);
 }
 vec2 eye_getScaledEyesPoint(vec2 originalPoint, vec2 scalingFactors,float scalingFactor,float leftEyeMask, float rightEyeMask) {
     if (scalingFactor == 0.0) {
         return originalPoint;
     }
     vec2 eyesScaledPoint = originalPoint;
     vec2 leftEyeScaledPoint = posV[1] + scalingFactors * (originalPoint - posV[1]);
     eyesScaledPoint = mix(eyesScaledPoint, leftEyeScaledPoint, leftEyeMask);
     vec2 rightEyeScaledPoint = posV[3] + scalingFactors * (originalPoint - posV[3]);
     eyesScaledPoint = mix(eyesScaledPoint, rightEyeScaledPoint, rightEyeMask);
     return mix(originalPoint, eyesScaledPoint, scalingFactor);
 }
 vec2 eye_getEyesDistanceFactors() {
     const float kEyesDistanceXRatio = 0.03662283339;
     return vec2(abs(kEyesDistanceXRatio * (posV[1] - posV[3]).x), 0.0);
 }
 vec2 eye_getEyesDistanceFactors2() {
     const float kEyesDistanceXRatio = 0.03662283339;
     return kEyesDistanceXRatio * vec2(posV[1].x - posV[3].x, posV[1].y - posV[3].y);
 }

 vec2 eye_getScaledEyesPoint2(vec2 originalPoint, float scalingFactor, float theta, float leftEyeMask, float rightEyeMask) {
     if (scalingFactor == 0.0) {
         return originalPoint;
     }
     float sin_t2 = sin(theta);
     float cos_t2 = cos(theta);
     vec2 v = originalPoint - posV[1];
     float d = dot(v, vec2(cos_t2, sin_t2));
     d = d * 0.3 * leftEyeMask * scalingFactor;
     vec2 left = d * vec2(cos_t2, sin_t2);

     v = originalPoint - posV[3];
     d = dot(v, vec2(cos_t2, sin_t2));
     d = d * 0.3 * rightEyeMask * scalingFactor;
     vec2 right = d * vec2(cos_t2, sin_t2);
     return originalPoint+left+right;
 }

 vec2 getEyeDiff() {
     //faceIndex;
     float leftEyeFineMask     = eye_getLeftEyeFineMask(vec2(1.0, 0.6), 0.9, 1.1);
     float rightEyeFineMask    = eye_getRightEyeFineMask(vec2(1.0, 0.6), 0.9, 1.1);
     float leftEyeCoarseMask   = eye_getLeftEyeFineMask(vec2(0.8, 0.7), 0.7, 1.2);
     float rightEyeCoarseMask  = eye_getRightEyeFineMask(vec2(0.8, 0.7), 0.7, 1.2);
     vec2 displacedEyesPoint   = textureCoordinateFix;
     //textureCoordinate = vec2(leftEyeFineMask+rightEyeFineMask);
     //gl_Position = vec4(position.x*2.0-1.0, position.y*2.0-1.0, 0.0, 1.0);
     //return;
     // width
     vec2 kEyesWidthFactors = vec2(1.3*cos_t, -1.3*sin_t);
     //vec3 kEyesWidthFactors = vec3(1.3, 1.0, 1.0);
     displacedEyesPoint = eye_getScaledEyesPoint2(displacedEyesPoint, paramV[1], angles.z, leftEyeFineMask, rightEyeFineMask);


     // height
     displacedEyesPoint = eye_getScaledEyesPoint2(displacedEyesPoint, paramV[2], kPi/2.0 + angles.z, leftEyeFineMask, rightEyeFineMask);

     // all size
     const vec2 kEyesSizeFactors = vec2(1.4, 1.4);
     displacedEyesPoint = eye_getScaledEyesPoint(displacedEyesPoint, kEyesSizeFactors, paramV[0], leftEyeFineMask, rightEyeFineMask);

     // rotate
     vec2 leftEye2DTiltPoint = posV[1] + rotate2DPoint(displacedEyesPoint - posV[1], -paramV[3]);
     displacedEyesPoint = mix(displacedEyesPoint, leftEye2DTiltPoint, leftEyeCoarseMask);
     vec2 rightEye2DTiltPoint = posV[3] + rotate2DPoint(displacedEyesPoint - posV[3], paramV[3]);
     displacedEyesPoint = mix(displacedEyesPoint, rightEye2DTiltPoint, rightEyeCoarseMask);


     // distance
     vec2 eyesDistanceFactors = eye_getEyesDistanceFactors2();
     vec2 leftEyeDistancePoint = displacedEyesPoint + paramV[4] * eyesDistanceFactors;
     displacedEyesPoint = mix(displacedEyesPoint, leftEyeDistancePoint, leftEyeCoarseMask);
     vec2 rightEyeDistancePoint = displacedEyesPoint - paramV[4] * eyesDistanceFactors;
     displacedEyesPoint = mix(displacedEyesPoint, rightEyeDistancePoint, rightEyeCoarseMask);


     vec2 diffPosition = displacedEyesPoint - textureCoordinateFix;
     return diffPosition;
 }

 float nose_getNoseMask(vec2 sigmaScaling, float a1, float a2) {
     return getEllipseMask(sigmaScaling, textureCoordinateFix, posV[4], posV[5], a1, a2);
 }
 float nose_getNoseTipFineMask() {
     return getEllipseMask(vec2(2.0, 2.0), textureCoordinateFix, posV[6], posV[7], 0.4, 0.8);
 }
 float nose_getNoseTipCoarseMask() {
     return getEllipseMask(vec2(1.0, 1.2), textureCoordinateFix, posV[8], posV[7], 0.5, 1.2);
 }
 float nose_getNoseVerticalAxisMask() {
     return getEllipseMask(vec2(2.0, 1.0), textureCoordinateFix, posV[4], posV[5], 0.5, 1.2);
 }
 float nose_getNoseVerticalAxisFineMask() {
     return getEllipseMask(vec2(2.5, 0.8), textureCoordinateFix, posV[4], posV[5], 0.5, 1.2);
 }
 float nose_getLipsMask(vec2 sigmaScaling, float a1, float a2) {
     return getEllipseMask(sigmaScaling, textureCoordinateFix, posV[9], posV[10], a1, a2);
 }

 vec2 nose_getScaledNosePoint(vec2 originalPoint, vec2 center, float scale) {
     vec2 v = originalPoint - center;
     float d = dot(v, vec2(cos_t, sin_t));
     d = d * scale;
     vec2 dv = d * vec2(cos_t, sin_t);

     return originalPoint+dv;
 }

 vec2 getNoseDiff() {
     float noseNegativeRegionsMask = (1.0 - eye_getLeftEyeFineMask(vec2(0.8), 0.5, 1.2)) * (1.0 - eye_getRightEyeFineMask(vec2(0.8), 0.5, 1.2)) * (1.0 - nose_getLipsMask(vec2(1.8), 0.5, 1.0));
     float noseMask = nose_getNoseMask(vec2(1.0, 1.5), 0.5, 1.2) * noseNegativeRegionsMask;
     float noseTipFineMask = nose_getNoseTipFineMask() * noseNegativeRegionsMask;
     float noseTipCoarseMask = nose_getNoseTipCoarseMask() * noseNegativeRegionsMask;
     float noseVertivalAxisMask = nose_getNoseVerticalAxisMask() * noseNegativeRegionsMask;
     float noseVertivalAxisFineMask = nose_getNoseVerticalAxisFineMask() * noseNegativeRegionsMask;
     vec2 displacedNosePoint = textureCoordinateFix;

     vec2 noseOutlineWidthPoint = nose_getScaledNosePoint(displacedNosePoint, posV[4], 0.3);
     displacedNosePoint = mix(displacedNosePoint, noseOutlineWidthPoint, paramV[6] * noseTipCoarseMask*(1.0-noseVertivalAxisFineMask));

     vec2 noseBridgeWidthPoint = nose_getScaledNosePoint(displacedNosePoint, posV[4], 0.3);
     displacedNosePoint = mix(displacedNosePoint, noseBridgeWidthPoint, paramV[7] * noseVertivalAxisMask);


     vec2 noseSizePoint = nose_getScaledNosePoint(displacedNosePoint, posV[4], 0.3);
     displacedNosePoint = mix(displacedNosePoint, noseSizePoint, paramV[5] * noseMask);


     const float kNoseLiftingYRatio = 0.05518821053;
     vec2 noseLiftingVector = vec2(0.0, -kNoseLiftingYRatio * posV[5].y);
     noseLiftingVector = vec2(dot(noseLiftingVector, vec2(cos_t, -sin_t)), dot(noseLiftingVector, vec2(sin_t, cos_t)));
     vec2 nosePositionPoint = displacedNosePoint + noseLiftingVector;
     displacedNosePoint = mix(displacedNosePoint, nosePositionPoint, paramV[8] * noseMask);

     vec2 noseTipPoint = posV[6] + vec2(0.4) * (displacedNosePoint - posV[6]);
     displacedNosePoint = mix(displacedNosePoint, noseTipPoint, paramV[9] * noseTipFineMask);


     vec2 diffPosition = displacedNosePoint - textureCoordinateFix;
     return diffPosition;
 }
 vec2 lips_getScaledLipsPoint(in vec2 originalPoint, in vec2 scalingFactors, in float scalingCoefficient, in float lipsMask) {
     if (scalingCoefficient == 0.0) {
         return originalPoint;
     }
     vec2 lipsScaledPoint = posV[9] + scalingFactors * (originalPoint - posV[9]);
     return mix(originalPoint, lipsScaledPoint, scalingCoefficient * lipsMask);
 }

 vec2 lips_getScaledLipsPoint2(vec2 originalPoint, vec2 center, float theta, float scale) {
     float sin_t = sin(theta);
     float cos_t = cos(theta);
     vec2 v = originalPoint - center;
     float d = dot(v, vec2(cos_t, sin_t));
     d = d * scale;
     vec2 dv = d * vec2(cos_t, sin_t);

     return originalPoint+dv;
 }
 vec2 lips_getTransformYFactors() {
     const float kDistanceYRatio = 0.1;
     return kDistanceYRatio * (posV[4] - posV[9]) / 2.0;
 }
 float lips_face_getFaceMaskLeft2(vec2 point) {
     return getEllipseMask2(vec2(1.0, 2.0), point, posV[16], posV[15], 0.8, 1.2, posV[17].x);
 }
 float lips_face_getFaceMaskRight2(vec2 point) {
     return getEllipseMask2(vec2(1.0, 2.0), point, posV[19], posV[18], 0.8, 1.2, posV[17].y);
 }

 vec2 getLipsDiff() {
     const float kScalingFactor = 1.3;
     vec2 originalPosition = textureCoordinateFix;
     vec2 faceFrameCoords = originalPosition;
     float teethRegionWeight = 0.0;
     float noseMask = nose_getNoseMask(vec2(1.0, 1.2), 0.8, 1.0);
     float faceCoarseMaskLeft2   = lips_face_getFaceMaskLeft2(textureCoordinateFix);
     float faceCoarseMaskRight2   = lips_face_getFaceMaskRight2(textureCoordinateFix);
     float lipsNegativeMask = (1.0 - teethRegionWeight) * (1.0 - noseMask) * (1.0 - faceCoarseMaskLeft2) * (1.0 - faceCoarseMaskRight2);
     float lipsMask = nose_getLipsMask(vec2(1.2, 1.2), 0.8, 1.1) * lipsNegativeMask;
//     lipsMask = lips_face_getFaceMaskLeft2(textureCoordinateFix) + lips_face_getFaceMaskRight2(textureCoordinateFix);
     float lipsMask1 = nose_getLipsMask(vec2(1.2, 1.1), 0.8, 0.9) * lipsNegativeMask;
     float lipsMask2 = nose_getLipsMask(vec2(1.2, 0.8), 1.0, 1.2) * lipsNegativeMask;
     vec2 displacedLipsPoint = textureCoordinateFix;

     const vec2 kLipsWidthFactors = vec2(kScalingFactor, 1.0);
     displacedLipsPoint = mix(displacedLipsPoint, lips_getScaledLipsPoint2(displacedLipsPoint, posV[9], angles.z, 0.3), paramV[11] * lipsMask1);


     const vec2 kLipsHeightFactors = vec2(1.0, kScalingFactor);
     displacedLipsPoint = mix(displacedLipsPoint, lips_getScaledLipsPoint2(displacedLipsPoint, posV[9], kPi/2.0 + angles.z, 0.3), paramV[12] * lipsMask);


     const vec2 kLipsSizeFactors = vec2(kScalingFactor, kScalingFactor);
     displacedLipsPoint = lips_getScaledLipsPoint(displacedLipsPoint, kLipsSizeFactors, paramV[10], lipsMask);


     vec2 lipsYFactors = lips_getTransformYFactors();
     vec2 lipsTransformYPoint = displacedLipsPoint + lipsYFactors;
     displacedLipsPoint = mix(displacedLipsPoint, lipsTransformYPoint, paramV[13] * lipsMask2);

     vec2 diffPosition = displacedLipsPoint - textureCoordinateFix;
     return diffPosition;
 }

 float brow_getLeftEyebrowMask() {
     return getEllipseMask(vec2(1.4, 0.8), textureCoordinateFix, posV[11], posV[12], 0.7, 1.0);
 }

 float brow_getRightEyebrowMask() {
     return getEllipseMask(vec2(1.4, 0.8), textureCoordinateFix, posV[13], posV[14], 0.7, 1.0);
 }

 float brow_getLeftEyebrowMask2() {
     return getEllipseMask(vec2(1.0, 0.7), textureCoordinateFix, posV[11], posV[12], 0.6, 1.0);
 }

 float brow_getRightEyebrowMask2() {
     return getEllipseMask(vec2(1.0, 0.7), textureCoordinateFix, posV[13], posV[14], 0.6, 1.0);
 }
 vec2 brow_getEyebrowDistanceFactors() {
     const float kEyesDistanceXRatio = 0.03662283339;
     return kEyesDistanceXRatio * vec2(posV[11].x - posV[13].x, posV[11].y - posV[13].y);
 }
 vec2 brow_getEyebrowHeightFactors() {
     const float kEyesDistanceXRatio = 0.14;
     return kEyesDistanceXRatio * (posV[11] + posV[13] - posV[1] - posV[3]) / 2.0;
 }
 vec2 getEyebrowDiff() {
     float _noseNegativeMask    = 1.0 - nose_getNoseMask(vec2(3.0, 0.9), 0.7, 0.9);
     float _leftEyeNegativeMask = 1.0 - eye_getLeftEyeFineMask(vec2(1.5, 1.5), 0.6, 0.9);
     float _rightEyeNegativeMask = 1.0 - eye_getRightEyeFineMask(vec2(1.5, 1.5), 0.6, 0.9);
     float leftEyebrowMask   = brow_getLeftEyebrowMask() * _leftEyeNegativeMask * _noseNegativeMask;
     float rightEyebrowMask  = brow_getRightEyebrowMask() * _rightEyeNegativeMask * _noseNegativeMask;
     float leftEyebrowMask2   = brow_getLeftEyebrowMask2() * _leftEyeNegativeMask * _noseNegativeMask;
     float rightEyebrowMask2  = brow_getRightEyebrowMask2() * _rightEyeNegativeMask * _noseNegativeMask;
     vec2 displacedEyebrowPoint   = textureCoordinateFix;

     // rotate

     vec2 leftEyebrow2DTiltPoint = posV[11] + rotate2DPoint(displacedEyebrowPoint - posV[11], -paramV[16]);
     displacedEyebrowPoint = mix(displacedEyebrowPoint, leftEyebrow2DTiltPoint, leftEyebrowMask2);
     vec2 rightEyebrow2DTiltPoint = posV[13] + rotate2DPoint(displacedEyebrowPoint - posV[13], paramV[16]);
     displacedEyebrowPoint = mix(displacedEyebrowPoint, rightEyebrow2DTiltPoint, rightEyebrowMask2);


     // distance
     vec2 eyebrowDistanceFactors = brow_getEyebrowDistanceFactors();
     vec2 leftEyeBrowDistancePoint = displacedEyebrowPoint + paramV[15] * eyebrowDistanceFactors;
     displacedEyebrowPoint = mix(displacedEyebrowPoint, leftEyeBrowDistancePoint, leftEyebrowMask);
     vec2 rightEyebrowDistancePoint = displacedEyebrowPoint - paramV[15] * eyebrowDistanceFactors;
     displacedEyebrowPoint = mix(displacedEyebrowPoint, rightEyebrowDistancePoint, rightEyebrowMask);


     vec2 eyebrowHeightFactors = brow_getEyebrowHeightFactors();
     vec2 leftEyeBrowHeightPoint = displacedEyebrowPoint + paramV[14] * eyebrowHeightFactors;
     displacedEyebrowPoint = mix(displacedEyebrowPoint, leftEyeBrowHeightPoint, leftEyebrowMask);
     vec2 rightEyebrowHeightPoint = displacedEyebrowPoint + paramV[14] * eyebrowHeightFactors;
     displacedEyebrowPoint = mix(displacedEyebrowPoint, rightEyebrowHeightPoint, rightEyebrowMask);

     vec2 diffPosition = displacedEyebrowPoint - textureCoordinateFix;
     return diffPosition;
 }

 void main(){
     vec2 trueDiff = vec2(0.0, 0.0);

     vec4 texDiff = texture2D(inputImageTexture, textureCoordinate);
     trueDiff = (texDiff.xy * 255.0 + texDiff.zw) / 127.5 - 1.0;   //diff of raw
     if (trueDiff.x < -0.9 && trueDiff.y < -0.9){
         trueDiff = vec2(0.0, 0.0);
     }

     textureCoordinateFix = (textureCoordinate * size - 0.5) / (size - 1.0);
     vec2 diffPosition = getEyeDiff() + getNoseDiff() + getLipsDiff() + getEyebrowDiff() + trueDiff;
     diffPosition = 0.5 * (diffPosition + 1.0) * 255.0;
     vec2 a = floor(diffPosition) / 255.0;
     vec2 b = fract(diffPosition);
     gl_FragColor = vec4(a,b);

 }