precision highp float;

//common
uniform float sin_t;
uniform float cos_t;
uniform vec3 angles; // angles.x: pitch, angles.y: yaw, angles.z: roll
uniform vec2 size;
varying vec2 textureCoordinate;

vec2 textureCoordinateFix; //used for fix error
uniform vec2 posV[31];
uniform float paramV[5];

const float kLipsEdgeWrinkleSizeToLipsSizeRatio = 0.23;
const float kPi = 3.14159265358979;
const float kLipsEdgeWrinklesRotationAngle = 0.025 * kPi;

uniform sampler2D inputImageTexture;

//common func
float getEllipseMask(in vec2 sigmaScaling, in vec2 point,in vec2 center, in vec2 boundingSize, float a1, float a2) {
    // un-normalization
    //boundingSize = boundingSize * size;
    //point = point * size;
    //center = center * size;

    vec2 translate = (point - center);
    // rotate and scale
    vec2 v = vec2(dot(translate, vec2(cos_t, sin_t)), dot(translate, vec2(-sin_t, cos_t))) * sigmaScaling;

    float d2 = sqrt(dot(v / boundingSize, v / boundingSize));

    return 1.0 - smoothstep(a1, a2, d2);
}

float getEllipseMask2(in vec2 sigmaScaling, in vec2 point,in vec2 center, in vec2 boundingSize, float a1, float a2, float ang) {
    // un-normalization
    //boundingSize = boundingSize * size;
    //point = point * size;
    //center = center * size;

    float cos_t = cos(ang);
    float sin_t = sin(ang);
    vec2 translate = (point - center);
    // rotate and scale
    vec2 v = vec2(dot(translate, vec2(cos_t, sin_t)), dot(translate, vec2(-sin_t, cos_t))) * sigmaScaling;

    float d2 = sqrt(dot(v / boundingSize, v / boundingSize));

    return 1.0 - smoothstep(a1, a2, d2);
}
vec2 rotate2DPoint(vec2 point, float angle) {
    return vec2(cos(angle) * point.x - sin(angle) * point.y, sin(angle) * point.x + cos(angle) * point.y);
}

float face_getLeftEyeMask() {
    return getEllipseMask(vec2(1.5, 0.9), textureCoordinateFix, posV[1], posV[0], 0.5, 1.2);
}
float face_getRightEyeMask() {
    return getEllipseMask(vec2(1.5, 0.9), textureCoordinateFix, posV[3], posV[2], 0.5, 1.2);
}
float face_getLeftEyebrowMask() {
    return getEllipseMask(vec2(1.1, 1.0), textureCoordinateFix, posV[4], posV[5], 0.5, 1.0);
}
float face_getRightEyebrowMask() {
    return getEllipseMask(vec2(1.1, 1.0), textureCoordinateFix, posV[6], posV[7], 0.5, 1.0);
}
float face_getLipsMask() {
    return getEllipseMask(vec2(1.5, 1.5), textureCoordinateFix, posV[10], posV[11], 0.5, 1.2);
}
float face_getNoseMask() {
    return getEllipseMask(vec2(1.1, 1.0), textureCoordinateFix, posV[8], posV[9], 0.5, 1.2);
}
float face_getFaceMask() {
    return getEllipseMask(vec2(1.5, 1.2), textureCoordinateFix, posV[15], posV[14], 0.7, 0.9);
}
float face_getFaceMaskLeft2(vec2 point) {
    return getEllipseMask2(vec2(4.0, 1.4), point, posV[17], posV[16], 0.5, 2.0, posV[18].x);
}
float face_getFaceMaskRight2(vec2 point) {
    return getEllipseMask2(vec2(4.0, 1.4), point, posV[20], posV[19], 0.5, 2.0, posV[18].y);
}
float face_getFaceMaskUpLeft2(vec2 point) {
    return getEllipseMask2(vec2(2.0, 1.4), point, posV[23], posV[22], 0.5, 0.8, posV[24].x);
}
float face_getFaceMaskUpRight2(vec2 point) {
    return getEllipseMask2(vec2(2.0, 1.4), point, posV[26], posV[25], 0.5, 0.8, posV[24].y);
}
vec2 face_getScaledFacePoint2(vec2 originalPoint, float scalingFactor, float faceCoarseMask) {
    if (scalingFactor == 0.0) {
        return originalPoint;
    }
    vec2 v = originalPoint - posV[15];
    float d = dot(v, vec2(cos_t, sin_t));
    d = d * 0.45 * faceCoarseMask * scalingFactor;
    vec2 sc = d * vec2(cos_t, sin_t);
    return originalPoint+sc;
}

vec2 getFaceWidthDiff() {
     vec2 diffPosition = vec2(0, 0);
     vec2 displacedFacePoint = textureCoordinateFix;


     float faceNegativeRegionsMask = (1.0 - 0.8*face_getLeftEyeMask()) * (1.0 - 0.8*face_getRightEyeMask()) * (1.0 - 0.8*face_getLeftEyebrowMask()) * (1.0 - 0.8*face_getRightEyebrowMask()) * (1.0 - 0.6*face_getLipsMask()) * (1.0 - 0.9*face_getNoseMask());
     float faceCoarseMask   = face_getFaceMask();
     //width
     displacedFacePoint   = face_getScaledFacePoint2(displacedFacePoint, paramV[0], faceCoarseMask*faceNegativeRegionsMask);

     diffPosition = displacedFacePoint - textureCoordinateFix;

     vec2 displacedFacePoint1 = displacedFacePoint;
     for (int i = 0; i < 5; i++){   //4 times
         float thres = 1.2;
         if (i == 4) thres = 1.0;
         displacedFacePoint1 = mix(displacedFacePoint1, displacedFacePoint1 - (posV[21] - posV[17])* 0.1, face_getFaceMaskLeft2(displacedFacePoint1) * paramV[1] * thres);
         displacedFacePoint1 = mix(displacedFacePoint1, displacedFacePoint1 - (posV[21] - posV[20])* 0.1, face_getFaceMaskRight2(displacedFacePoint1) * paramV[1] * thres);
     }
     vec2 displacedFacePoint2 = displacedFacePoint;
     for (int i = 0; i < 3; i++){
         displacedFacePoint2 = mix(displacedFacePoint2, displacedFacePoint2 - (posV[21] - posV[23])* 0.1, face_getFaceMaskUpLeft2(displacedFacePoint2) * paramV[1]);
         displacedFacePoint2 = mix(displacedFacePoint2, displacedFacePoint2 - (posV[21] - posV[26])* 0.1, face_getFaceMaskUpRight2(displacedFacePoint2) * paramV[1]);
     }
     diffPosition = displacedFacePoint1 - displacedFacePoint + displacedFacePoint2 - displacedFacePoint + diffPosition;


     return diffPosition;
 }


float jaw_getLipsMask() {
    return getEllipseMask(vec2(1.4, 1.0), textureCoordinateFix, posV[10], posV[11], 0.5, 1.2);
}
float jaw_getNoseMask() {
    return getEllipseMask(vec2(1.0, 1.5), textureCoordinateFix, posV[8], posV[9], 0.5, 1.2);
}
float jaw_getJawMask() {
    return getEllipseMask(vec2(1.8, 1.0), textureCoordinateFix, posV[27], posV[28], 0.5, 1.2);
}
vec2 getJawDiff() {
    float lipsMask = jaw_getLipsMask();
    float noseMask = jaw_getNoseMask();
    float jawMask = jaw_getJawMask() * (1.0 - 0.9 * lipsMask) * (1.0 - 0.9 * noseMask);

    vec2 displacedPoint = textureCoordinateFix;

    const float kJawLiftingYRatio = 0.1;
    vec2 jawLiftingVector = vec2(0.0, kJawLiftingYRatio * posV[28].y);
    jawLiftingVector = vec2(dot(jawLiftingVector, vec2(cos_t, -sin_t)), dot(jawLiftingVector, vec2(sin_t, cos_t)));
    vec2 jawPositionPoint = displacedPoint + jawLiftingVector;
    vec2 displacedPoint2 = mix(displacedPoint, jawPositionPoint, paramV[2] * jawMask);

    const vec2 kScaleFactors = vec2(1.1, 1.1);
    displacedPoint = posV[27] + kScaleFactors * (textureCoordinateFix - posV[27]);
    displacedPoint = mix(textureCoordinateFix, displacedPoint, paramV[2] * jawMask);

    displacedPoint = displacedPoint * 0.65 + displacedPoint2 * 0.35;
    vec2 diffPosition = displacedPoint - textureCoordinateFix;
    return diffPosition;
}

float head_getLeftEyebrowMask() {
    return getEllipseMask(vec2(1.0, 0.8), textureCoordinateFix, posV[4], posV[5], 0.6, 1.0);
}
float head_getRightEyebrowMask() {
    return getEllipseMask(vec2(1.0, 0.8), textureCoordinateFix, posV[6], posV[7], 0.6, 1.0);
}
float head_getForeheadMask() {
    return getEllipseMask(vec2(1.9, 1.6), textureCoordinateFix, posV[29], posV[30], 0.2, 1.4);
}

vec2 getForeHeadDiff() {
    float foreheadMask    = head_getForeheadMask() * (1.0 - head_getLeftEyebrowMask()) * (1.0 - head_getRightEyebrowMask());
    vec2 displacedForeheadPoint   = textureCoordinateFix;

    const float kForeheadLiftingYRatio = 0.04;
    vec2 foreheadLiftingVector = vec2(0.0, -kForeheadLiftingYRatio * posV[30].y);
    foreheadLiftingVector = vec2(dot(foreheadLiftingVector, vec2(cos_t, -sin_t)), dot(foreheadLiftingVector, vec2(sin_t, cos_t)));
    vec2 foreheadPositionPoint = displacedForeheadPoint + foreheadLiftingVector;
    displacedForeheadPoint = mix(displacedForeheadPoint, foreheadPositionPoint, foreheadMask * paramV[4]);

    vec2 diffPosition = displacedForeheadPoint - textureCoordinateFix;
    return diffPosition;
}

vec2 wrinkle_rotate2DPointAroundCenter(in vec2 point, in vec2 center, in float angle) {
    return center + rotate2DPoint(point - center, angle);
}
vec2 wrinkle_getRotatedLipsEdgeWrinkle(in vec2 originalPoint, in int isLeftSide) {
    vec2 wrinkleSize = 1.1*vec2(kLipsEdgeWrinkleSizeToLipsSizeRatio * posV[11].x, 0.6*posV[11].y);
    vec2 lipsEdge = isLeftSide == 1 ? posV[12] : posV[13];
    float directionSign = isLeftSide == 1 ? 1.0 : -1.0;
    vec2 wrinkleCenter = lipsEdge + directionSign * 0.055 * (posV[13] - posV[12]);

    float wrinkleMask = getEllipseMask(vec2(0.8, 0.5), textureCoordinateFix, wrinkleCenter,wrinkleSize, 0.5, 1.2);
    vec2 wrinkleRotationCenter = lipsEdge + directionSign * 0.2 * (posV[13] - posV[12]);
    vec2 rotated2DPoint = wrinkle_rotate2DPointAroundCenter(originalPoint, wrinkleRotationCenter, - directionSign * wrinkleMask * kLipsEdgeWrinklesRotationAngle);
    return rotated2DPoint;
}
vec2 wrinkle_getRotatedLipsLeftEdgeWrinkle(in vec2 originalPoint) {
    return wrinkle_getRotatedLipsEdgeWrinkle(originalPoint, 1);
}
vec2 wrinkle_getRotatedLipsRightEdgeWrinkle(in vec2 originalPoint) {
    return wrinkle_getRotatedLipsEdgeWrinkle(originalPoint, 0);
}

float wrinkle_getNoseMask() {
    return getEllipseMask(vec2(1.2, 1.5), textureCoordinateFix, posV[8].xy, posV[9], 0.8, 1.0);
}
vec2 getWrinkleDiff() {

    vec2 displacedPoint = textureCoordinateFix;
    float _noseMask = wrinkle_getNoseMask() * 0.9;
    displacedPoint = mix(wrinkle_getRotatedLipsLeftEdgeWrinkle(displacedPoint), displacedPoint, _noseMask);
    displacedPoint = mix(wrinkle_getRotatedLipsRightEdgeWrinkle(displacedPoint), displacedPoint, _noseMask);
    vec2 displacedPosition = textureCoordinateFix + paramV[3] * (displacedPoint - textureCoordinateFix);
    vec2 diffPosition = displacedPosition - textureCoordinateFix;
    return diffPosition;
}

void main(){
    vec2 trueDiff = vec2(0.0, 0.0);

    vec4 texDiff = texture2D(inputImageTexture, textureCoordinate);
    trueDiff = (texDiff.xy * 255.0 + texDiff.zw) / 127.5 - 1.0;   //diff of raw
    if (trueDiff.x < -0.9 && trueDiff.y < -0.9){
        trueDiff = vec2(0.0, 0.0);
    }

    textureCoordinateFix = (textureCoordinate * size - 0.5) / (size - 1.0);
    vec2 diffPosition = getJawDiff() + getFaceWidthDiff() + getForeHeadDiff() + getWrinkleDiff() + trueDiff;
    diffPosition = 0.5 * (diffPosition + 1.0) * 255.0;
    vec2 a = floor(diffPosition) / 255.0;
    vec2 b = fract(diffPosition);
    gl_FragColor = vec4(a,b);


}
